// Scene_Life03 v1.1 by MacGruber
// Scene controller plugin for scene "Life03", which is part of the Life plugin.

using System.Linq;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class Scene_Life03 : MVRScript
    {	
		private JSONStorable myBreathing;
		private JSONStorable myDriverBreathing;
		private JSONStorable myDriverThrust;
		private JSONStorableFloat myBreathingIntensity;
		private JSONStorableFloat myBreathingMouthCloseTime;
		private JSONStorableFloat myDriverThrustPower;
		private JSONStorableFloat myDriverThrustMultiplier;
		private JSONStorableFloat myDriverThrustOffset;
		private JSONStorableFloat myDriverThrustRhythmDamping;
		
		private float intensityChangeClock = 0.0f;
		private float intensityChangeDuration = 0.0f;
		
        public override void Init()
		{
			if (containingAtom.type != "Person")
			{
				SuperController.LogError($"Requires atom of type person, instead selected: {containingAtom.type}");
				return;
			}

			myBreathing = FindPlugin("MacGruber.Breathing");
			myDriverBreathing = FindPlugin("MacGruber.DriverBreathing");
			myDriverThrust = FindPlugin("MacGruber.DriverThrust");			
			if (myBreathing == null)
			{
				SuperController.LogError($"Scene_Life03 requires MacGruber.Breathing plugin!");
				return;
			}
			if (myDriverBreathing == null)
			{
				SuperController.LogError($"Scene_Life03 requires MacGruber.DriverBreathing plugin!");
				return;
			}
			if (myDriverThrust == null)
			{
				SuperController.LogError($"Scene_Life03 requires MacGruber.DriverThrust plugin!");
				return;
			}
			
			myBreathingIntensity = myBreathing.GetFloatJSONParam("Intensity");
			myBreathingMouthCloseTime = myDriverBreathing.GetFloatJSONParam("Mouth Close Time");
			myDriverThrustPower = myDriverThrust.GetFloatJSONParam("Power");
			myDriverThrustMultiplier = myDriverThrust.GetFloatJSONParam("Multiplier");
			myDriverThrustOffset = myDriverThrust.GetFloatJSONParam("Offset");
			myDriverThrustRhythmDamping = myDriverThrust.GetFloatJSONParam("Rhythm Damping");
			
			OnChooseBreath();
		}
		
		private JSONStorable FindPlugin(string fullClassName)
		{
			List<string> names = containingAtom.GetStorableIDs();
			string pluginName = names.Find(s => s.StartsWith("plugin#") && s.EndsWith(fullClassName));
			JSONStorable plugin = containingAtom.GetStorableByID(pluginName);
			return plugin;
		}
		
        private void Update()
		{
			intensityChangeClock += Time.deltaTime;
			
			if (intensityIndex >= 0 && intensityIndex < intensityData.Length)
			{				
				IntensityData data = intensityData[intensityIndex];
				myDriverThrustPower.val = Mathf.Lerp(myDriverThrustPower.val, data.thrustPower, Mathf.Min(1.0f, Time.deltaTime));
			}
		}
		
		private void OnChooseBreath() // Listening for a UnityMessage send by the Breathing plugin.
		{	
			if (intensityChangeClock < intensityChangeDuration)
				return;
			
			if (intensityIndex < 0 || intensityIndex >= intensityData.Length)
			{
				intensityIndex = 0;
			}
			else if (intensityData[intensityIndex].choices != null)
			{
				int choice = UnityEngine.Random.Range(0, intensityData[intensityIndex].choices.Length);
				intensityIndex = intensityData[intensityIndex].choices[choice];
			}
			else
			{
				int choice = UnityEngine.Random.Range(0, intensityData.Length);
				intensityIndex = choice;
			}
			
			//SuperController.LogMessage("OnChooseBreath State="+intensityIndex);

			IntensityData data = intensityData[intensityIndex];
			myBreathingIntensity.val = data.breathingIntensity;
			myBreathingMouthCloseTime.val = data.breathingMouthCloseTime;
			myDriverThrustMultiplier.val = data.thrustMultiplier;
			myDriverThrustOffset.val = data.thrustOffset;
			myDriverThrustRhythmDamping.val = data.thrustRhythmDamping;
			
			if (data.queueOrgasm)
				myBreathing.CallAction("QueueOrgasm");
			
			intensityChangeClock = 0.0f;
			intensityChangeDuration = UnityEngine.Random.Range(data.minDuration, data.maxDuration);
        }
		
		
		private class IntensityData
		{
			public float breathingIntensity = 0.5f;
			public float breathingMouthCloseTime = 0.2f;
			public float thrustPower = 0.5f;
			public float thrustMultiplier = 1.0f;
			public float thrustOffset = -0.2f;		
			public float thrustRhythmDamping = 0.5f;
			public float minDuration = 15.0f;
			public float maxDuration = 25.0f;
			public bool queueOrgasm = false;
			public int[] choices = null;
		}
		
		private int intensityIndex = -1;
		private IntensityData[] intensityData = new IntensityData[] {
			new IntensityData() { // 0
				breathingIntensity = 0.2f,
				thrustPower = 0.0f,
				minDuration = 3.0f,
				maxDuration = 5.0f,
				choices = new int[]{1,1,2}
			},
			new IntensityData() { // 1
				breathingIntensity = 0.5f,
				thrustPower = 0.7f,
				minDuration = 10.0f,
				maxDuration = 15.0f,
				choices = new int[]{2,2,3}
			},
			new IntensityData() { // 2
				breathingIntensity = 0.75f,
				thrustPower = 0.9f,
				thrustOffset = -0.2f,
				choices = new int[]{1,3,3,4}
			},
			new IntensityData() { // 3
				breathingIntensity = 0.8f,
				thrustPower = 1.0f,
				thrustMultiplier = 2.0f,
				thrustOffset = 0.0f,
				thrustRhythmDamping = 0.9f,
				choices = new int[]{2,4,4}		
			},
			new IntensityData() { // 4
				breathingIntensity = 1.0f,
				thrustPower = 1.0f,
				thrustMultiplier = 2.0f,
				thrustOffset = 0.0f,
				thrustRhythmDamping = 0.9f,
				minDuration = 10.0f,
				maxDuration = 15.0f,
				choices = new int[]{3,5}
			},
			new IntensityData() { // 5
				breathingIntensity = 0.2f,
				breathingMouthCloseTime = 0.05f,
				thrustPower = 1.0f,
				thrustMultiplier = 2.0f,
				thrustOffset = 0.0f,
				minDuration = 3.0f,
				maxDuration = 3.0f,
				queueOrgasm = true,
				choices = new int[]{0}
			}
		};
    }
}
